#include "stdafx.h"
#include "GameStatePhysics.h"
#include "GameStatePhysics/Utilities.h"


static ErrorStream		gMyErrorStream;
static UserAllocator*	gMyAllocator = NULL;
static bool				gPause = false;
static NxPhysicsSDK*	gPhysicsSDK = NULL;
static NxScene*			gScene = NULL;
static NxVec3			gDefaultGravity(0.0f, -9.81f, 0.0f);

#define TERRAIN_SIZE	33
#define TERRAIN_CHAOS	300.0f

#define TERRAIN_OFFSET	-20.0f
#define TERRAIN_WIDTH	20.0f
static TerrainData* gTerrainData = NULL;

// Render code
static NxVec3 Eye(50.0f, 50.0f, 50.0f);
static NxVec3 Dir(-0.6,-0.2,-0.7);
static NxVec3 N;
static int mx = 0;
static int my = 0;

static NxArray<NxU32> gTouchedTris;
static NxActor* gTerrain = NULL;
static NxTriangleMesh* terrainMesh = NULL;


class MyContactReport : public NxUserContactReport
{
public:
	virtual void  onContactNotify(NxContactPair& pair, NxU32 events)
	{
		// Iterate through contact points
		NxContactStreamIterator i(pair.stream);
		//user can call getNumPairs() here
		while(i.goNextPair())
		{
			//user can also call getShape() and getNumPatches() here
			while(i.goNextPatch())
			{
				//user can also call getPatchNormal() and getNumPoints() here
				const NxVec3& contactNormal = i.getPatchNormal();
				while(i.goNextPoint())
				{
					//user can also call getPoint() and getSeparation() here
					if(i.getSeparation()<0.0f)
					{
						const NxVec3& contactPoint = i.getPoint();

						NxU32 faceIndex = i.getFeatureIndex0();
						if(faceIndex==0xffffffff)	faceIndex = i.getFeatureIndex1();
						if(faceIndex!=0xffffffff)
						{
							gTouchedTris.pushBack(faceIndex);
						}
					}
				}
			}
		}
	}

}gMyContactReport;

static void CreatePhysicsTerrain()
{
	if(gTerrainData!=NULL)
	{
		delete gTerrainData;
		gTerrainData=NULL;		
	}

	gTerrainData = new TerrainData;
	gTerrainData->init(TERRAIN_SIZE, TERRAIN_OFFSET, TERRAIN_WIDTH, TERRAIN_CHAOS);

	// Build physical model
	NxTriangleMeshDesc terrainDesc;
	terrainDesc.numVertices					= gTerrainData->nbVerts;
	terrainDesc.numTriangles				= gTerrainData->nbFaces;
	terrainDesc.pointStrideBytes			= sizeof(NxVec3);
	terrainDesc.triangleStrideBytes			= 3*sizeof(NxU32);
	terrainDesc.points						= gTerrainData->verts;
	terrainDesc.triangles					= gTerrainData->faces;
	terrainDesc.flags						= 0;

	terrainDesc.heightFieldVerticalAxis		= NX_Y;
	terrainDesc.heightFieldVerticalExtent	= -1000.0f;

	bool status = InitCooking(gMyAllocator, &gMyErrorStream);
	if (!status) {
		printf("Unable to initialize the cooking library. Please make sure that you have correctly installed the latest version of the AGEIA PhysX SDK.");
		exit(1);
	}

	MemoryWriteBuffer buf;
	status = CookTriangleMesh(terrainDesc, buf);
	if (!status) {
		printf("Unable to cook a triangle mesh.");
		exit(1);
	}
	MemoryReadBuffer readBuffer(buf.data);
	terrainMesh = gPhysicsSDK->createTriangleMesh(readBuffer);

	//
	// Please note about the created Triangle Mesh, user needs to release it when no one uses it to save memory. It can be detected
	// by API "meshData->getReferenceCount() == 0". And, the release API is "gPhysicsSDK->releaseTriangleMesh(*meshData);"
	//

	NxTriangleMeshShapeDesc terrainShapeDesc;
	terrainShapeDesc.meshData				= terrainMesh;
	terrainShapeDesc.shapeFlags				= NX_SF_FEATURE_INDICES;

	NxActorDesc ActorDesc;
	ActorDesc.shapes.pushBack(&terrainShapeDesc);
	gTerrain = gScene->createActor(ActorDesc);
	gTerrain->userData = (void*)0;

	CloseCooking();
}

static void RenderTerrain()
{
	if(gTerrainData)
	{
		renderTerrainTriangles(*gTerrainData, gTouchedTris.size(), &gTouchedTris[0]);
		renderTerrain(*gTerrainData);
	}
}

void GameStatePhysics::destroyGameState(void)
{
	delete camera;
	camera = NULL;

	if(gTerrainData!=NULL)
	{
		delete gTerrainData;
		gTerrainData=NULL;
	}

	if(gPhysicsSDK != NULL)
	{
		if(gScene != NULL) gPhysicsSDK->releaseScene(*gScene);
		gScene = NULL;
		NxReleasePhysicsSDK(gPhysicsSDK);
		gPhysicsSDK = NULL;
	}

	if(gMyAllocator!=NULL)
	{
		delete gMyAllocator;
		gMyAllocator=NULL;
	}
}

bool GameStatePhysics::initNX(void)
{
	if (!gMyAllocator)
		gMyAllocator = new UserAllocator();

	// Initialize PhysicsSDK
	NxPhysicsSDKDesc desc;
	NxSDKCreateError errorCode = NXCE_NO_ERROR;
	gPhysicsSDK = NxCreatePhysicsSDK(NX_PHYSICS_SDK_VERSION, gMyAllocator, &gMyErrorStream, desc, &errorCode);
	if(gPhysicsSDK == NULL) 
	{
		printf("\nSDK create error (%d - %s).\nUnable to initialize the PhysX SDK, exiting.\n\n", errorCode, getNxSDKCreateError(errorCode));
		return false;
	}
#if SAMPLES_USE_VRD
	// The settings for the VRD host and port are found in SampleCommonCode/SamplesVRDSettings.h
	if (gPhysicsSDK->getFoundationSDK().getRemoteDebugger() && !gPhysicsSDK->getFoundationSDK().getRemoteDebugger()->isConnected())
		gPhysicsSDK->getFoundationSDK().getRemoteDebugger()->connect(SAMPLES_VRD_HOST, SAMPLES_VRD_PORT, SAMPLES_VRD_EVENTMASK);
#endif

	gPhysicsSDK->setParameter(NX_SKIN_WIDTH, 0.025f);

	// Create a scene
	NxSceneDesc sceneDesc;
	sceneDesc.gravity				= gDefaultGravity;
	sceneDesc.userContactReport		= &gMyContactReport;
	gScene = gPhysicsSDK->createScene(sceneDesc);
	if(gScene == NULL) 
	{
		printf("\nError: Unable to create a PhysX scene, exiting the sample.\n\n");
		return false;
	}

	NxMaterial * defaultMaterial = gScene->getMaterialFromIndex(0); 
	defaultMaterial->setRestitution(0.0f);
	defaultMaterial->setStaticFriction(0.5f);
	defaultMaterial->setDynamicFriction(0.5f);
	CreatePhysicsTerrain();

	return true;

}

void GameStatePhysics::initGameState(vector<vector<string> >& params)
{
	initNX();
	setDefaultParameterValues();
	setParamsAccordingToFileParams(params);

	glutSetWindowTitle("GameStatePhysics");

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

		// Setup default render states
	glClearColor(0.3f, 0.4f, 0.5f, 1.0);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_CULL_FACE);

	// Setup lighting
	glEnable(GL_LIGHTING);
	float AmbientColor[] = { 0.0f, 0.1f, 0.2f, 0.0f };
	glLightfv(GL_LIGHT0, GL_AMBIENT, AmbientColor);
	float DiffuseColor[] = { 1.0f, 1.0f, 1.0f, 0.0f };
	glLightfv(GL_LIGHT0, GL_DIFFUSE, DiffuseColor);
	float SpecularColor[] = { 0.0f, 0.0f, 0.0f, 0.0f };
	glLightfv(GL_LIGHT0, GL_SPECULAR, SpecularColor);
	float Position[] = { -10.0f, 100.0f, -4.0f, 1.0f };
	glLightfv(GL_LIGHT0, GL_POSITION, Position);
	glEnable(GL_LIGHT0);

	glutReshapeWindow(width, height);
}


void GameStatePhysics::processKey(unsigned char& key, int& x, int& y)
{
	printf("key pressed in GameStatePhysics: %i\n", key);
}

void GameStatePhysics::processKeyUp(unsigned char& key, int& x, int& y)
{
	printf("key released in GameStatePhysics: %i\n", key);
}

void GameStatePhysics::processSpecialKey(int& key, int& x, int& y)
{
	printf("Special key pressed in GameStatePhysics: %i\n", key);
}

void GameStatePhysics::processSpecialKeyUp(int& key, int& x, int& y)
{
	printf("Special key released in GameStatePhysics: %i\n", key);
}

void GameStatePhysics::processMouseClick(int& button, int& state, int& x, int& y)
{
}

void GameStatePhysics::processMousePassiveMovement(int& x, int& y)
{
}

void GameStatePhysics::processMouseMovement(int& x, int& y)
{
}
	
/*ResourceHandlers:*/


/*Rendering:*/
void GameStatePhysics::render(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glLoadIdentity();

	glutSwapBuffers();
}

void GameStatePhysics::update(void)
{
}

void GameStatePhysics::reshape(int w, int h)
{
	glViewport (0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (60, (GLfloat)w / (GLfloat)h, 1.0, 1000.0);
	glMatrixMode (GL_MODELVIEW);
}

/*
* Just for making the code better to read; attempts to retrieve parameter-values from the 
* given vector of parameter-tuples.
*/
void GameStatePhysics::setParamsAccordingToFileParams(vector<vector<string> >& params)
{
	if(params.size()>0)
	{
		for (int i = 0; i<(int)params.size(); i++)
		{
			vector<string> temp = params.at(i);
			string name = temp.at(0);
			string value = temp.at(1);
			if (value.compare("")==0) continue; //empty value -> skip

			if (name.compare("WindowWidth")==0) { width = atoi(value.c_str());continue; }
			if (name.compare("WindowHeight")==0) { height = atoi(value.c_str());continue; }
			if (name.compare("FullScreen")==0) { fullScreen = atoi(value.c_str())>0;continue; }
			//atoi = convertToInteger; atof = convertToFloat
		}
	}
}

/*
* (Re)Sets all parameter-values to default values.
*/
void GameStatePhysics::setDefaultParameterValues()
{
	width = 200;
	height = 200;
	fullScreen = false;
}