#ifndef __CUDA_STUFF
#define __CUDA_STUFF


#include "VPL.h"
#include "../GameStates/global.h"
#include <queue>
#include <list>
#include "../Timer.h"
#include "../../CUDA/CUDA_Structs.h"
#include "../../CUDA/CUDA_PushPull.h"
#include "../../CUDA/CUDA_Lightcuts.h"
#include <vector_types.h>

using namespace std;

struct Bounds
{
	Point A;
	Point B;
};

struct visualizationValues
{
	Bounds bounds;
	float x, y, z;
	float intensity;
};

class Cluster;
class Cluster
{
public:
	Cluster();
	~Cluster();
	Cluster(VPL* newCluster);
	Cluster(const Cluster* newCluster);
	
	VPL* representativeLight;
	Bounds bounds;

	bool sameOrientation;

	Cluster* leftChild;
	Cluster* rightChild;

	float intensity;

	int ID;
	int clusterNumber;

	//wird beim lightcut-refinen fuer jeden pixel neu gesetzt:
	double materialTerm;
	double geometryTerm;
	double visibilityTerm;
	double radianceUpperBound;
	double estimatedRadiance;
	double errorBound;

	bool operator==(const Cluster& v);
	
};
//vgl operator fuer: ein cluster hat kleinere upper bound als anderer
struct CompareClusterPtrByErrorBound
{
    bool operator()(Cluster* a1, Cluster* a2) 
	{
		return (*a1).errorBound < (*a2).errorBound;
		//return (*a1).radianceUpperBound < (*a2).radianceUpperBound;
	}
};


class ClusterPair
{
public:
	ClusterPair(Cluster* _A, Cluster* _B);
	~ClusterPair(void);
	Cluster* A;
	Cluster* B;
	float distance;
	
	bool operator<( const ClusterPair& rhs ) const
	{ 
		return distance > rhs.distance; 
	}
};
struct ComparePtrToA
{
    bool operator()(ClusterPair* a1, ClusterPair* a2) 
	{
		return *a1 < *a2;
	}
};


class KD_Node;
class KD_Node
{
public:
	KD_Node();
	KD_Node* leftChild;
	KD_Node* rightChild;
	KD_Node* parent;
	Cluster* thisNode;
	int separationAxis;
	bool isValid;
	Bounds maximumBoundingBox;
	float minimumIntensity;
};

struct CurrentBest
{
	KD_Node* bestMatch;
	float upperBound;
	float size;
};

class KD_Tree
{
public:
	KD_Node* root;
	bool isValid(Cluster* a);
	void buildTree(deque<Cluster*>* data);
	void recursiveFillTree(KD_Node* node, KD_Node* parent, deque<Cluster*>* data, int axis);
	void debugOutputTree(KD_Node* node);
	void debugOutputTreeToFile(KD_Node* node);
	void recursiveFileOutput(KD_Node* node, FILE* file);
	void recursiveDelete(KD_Node* node);
	KD_Node* updateTree(KD_Node* nodeA, KD_Node* nodeB, Cluster* merged);
	KD_Node* findClosestNeighbour(KD_Node* fixNode, KD_Node* currentSearchLoc);
	KD_Node* findNode(Cluster* searchThis, int ID, KD_Node* currentSearchLoc, int axis, bool* searchEnded);
	CurrentBest findClosestDown(KD_Node* fixNode, KD_Node* currentSearchLoc, CurrentBest currentBest);
	CurrentBest findClosestUp(KD_Node* fixNode, KD_Node* currentSearchLoc, KD_Node* previousSearchLoc, CurrentBest currentBest);
	void printValids(KD_Node* node);
	float sumOfIntensities(KD_Node* node, float value);

};


struct SortX : public std::binary_function<Cluster*, Cluster*, bool>
{
	bool operator()(Cluster* a, Cluster* b) const
	{                             
		return (a->representativeLight->position.x < b->representativeLight->position.x);
	};
};

struct SortY : public std::binary_function<Cluster*, Cluster*, bool>
{
	bool operator()(Cluster* a, Cluster* b) const
	{                             
		return (a->representativeLight->position.y < b->representativeLight->position.y);
	};
};

struct SortZ : public std::binary_function<Cluster*, Cluster*, bool>
{
	bool operator()(Cluster* a, Cluster* b) const
	{                             
		return (a->representativeLight->position.z < b->representativeLight->position.z);
	};
};


class CpuLightTree
{
public:
	CpuLightTree();
	~CpuLightTree();
	Cluster* root;
	deque<Cluster*>* lights;

	void recursiveDeleteCluster(Cluster* cluster);

	KD_Tree kd_tree;

	void addCluster(Cluster* cluster);


	void buildClusters(void);
	void buildClusters2(void);

	void debugOutput(Cluster* node);

	void debugOutputTreeToFile(Cluster* node);

	void recursiveFileOutput(Cluster* node, FILE* file);

	void debugOutputPriorityQueue(priority_queue<ClusterPair*, vector<ClusterPair*>, ComparePtrToA> someQueue);

};

class LightTree
{
private:
	int numLights;
	float* data;
	int nextFreeSlot;
public:
	LightTree(int numLights);//initializes structure
	~LightTree();
	void addVPL(VPL newVLP); //adds newVPL to the tree
	void addVPL(Point pos, Vector normal, float* color);
	void clusterLightTree(void);
	float* getLightTree(void); //returns the lightTree
	void debugOutput(void);
};

void doSomeDebugging();
float calculateClusterSize(Cluster* a, Cluster* b);
float calculateUpperBound(Cluster* a, Cluster* b);
Cluster* mergeClusters(Cluster* a, Cluster* b);
Bounds calcNewBounds(Bounds a, Bounds b);

void setCUDA_ClusterProperties(Cluster* original);




class LightCut
{
public:
	double sumIlluminationEstimates;
	double sumRadianceUpperBounds;
	Vector accColor;
	priority_queue<Cluster*, vector<Cluster*>, CompareClusterPtrByErrorBound> cut;

};


#endif